package utils

import (
	"archive/zip"
	"context"
	"io"
	"net/http"
	"os"
	"path/filepath"
	"runtime"
	"strings"

	"github.com/spaceuptech/helpers"
	"github.com/spaceuptech/space-api-go/utils"
)

// Unzip unzips a source file to a given destination
func Unzip(src string, dest string) error {

	r, err := zip.OpenReader(src)
	if err != nil {
		return err
	}
	defer utils.CloseTheCloser(r)

	for _, f := range r.File {

		// Store filename/path for returning and using later on
		fpath := filepath.Join(dest, f.Name)

		// Check for ZipSlip. More Info: http://bit.ly/2MsjAWE
		if !strings.HasPrefix(fpath, filepath.Clean(dest)+string(os.PathSeparator)) {
			return helpers.Logger.LogError(helpers.GetRequestID(context.TODO()), "Invalid file path provided", nil, nil)
		}

		if f.FileInfo().IsDir() {
			// Make Folder
			_ = os.MkdirAll(fpath, os.ModePerm)
			continue
		}

		// Make File
		if err = os.MkdirAll(filepath.Dir(fpath), os.ModePerm); err != nil {
			return err
		}

		outFile, err := os.OpenFile(fpath, os.O_WRONLY|os.O_CREATE|os.O_TRUNC, f.Mode())
		if err != nil {
			return err
		}

		rc, err := f.Open()
		if err != nil {
			return err
		}

		_, err = io.Copy(outFile, rc)

		// Close the file without defer to close before next iteration of loop
		utils.CloseTheCloser(outFile)
		utils.CloseTheCloser(rc)

		if err != nil {
			return err
		}
	}
	return nil
}

// DownloadFileFromURL downloads a file from url and stores it at a given destination
func DownloadFileFromURL(url, dest string) error {
	resp, err := http.Get(url)
	if err != nil {
		return err
	}
	defer utils.CloseTheCloser(resp.Body)

	// Create the file
	out, err := os.Create(dest)
	if err != nil {
		return err
	}
	defer utils.CloseTheCloser(out)

	// Write the body to file
	_, err = io.Copy(out, resp.Body)
	if err != nil {
		return err
	}
	return nil
}

// UserHomeDir returns the path of home directory of user
func UserHomeDir() string {
	if runtime.GOOS == "windows" {
		home := os.Getenv("HOMEDRIVE") + os.Getenv("HOMEPATH")
		if home == "" {
			home = os.Getenv("USERPROFILE")
		}
		return home
	}
	return os.Getenv("HOME")
}
